import sqlalchemy as sa
from sqlalchemy.schema import MetaData, Column, Table, ForeignKey
from sqlalchemy.types import Integer, String

metadata = MetaData()

ontologies = Table(
    'ontologies', metadata,
    Column('id', Integer, primary_key=True),
    Column('ontology', String, nullable=False),
    Column('class_id', String, nullable=False),
    Column('cui', String),
    Column('preferred_label', String),
    Column('definitions', String),
    Column('obsolete', String)
)

parents = Table(
    'parents', metadata,
    Column('id', Integer, primary_key=True),
    Column('ontology_id', Integer, ForeignKey('ontologies.id')),
    Column('parent', Integer, ForeignKey('ontologies.id'))
)

synonyms = Table(
    'synonyms', metadata,
    Column('id', Integer, primary_key=True),
    Column('ontology_id', Integer, ForeignKey('ontologies.id'), index=True),
    Column('synonym', String, index=True)
)

semantic_types = Table(
    'semantic_types', metadata,
    Column('id', Integer, primary_key=True),
    Column('ontology_id', Integer, ForeignKey('ontologies.id'), index=True),
    Column('semantic_type', String, index=True)
)


def create_db(url):
    engine = sa.create_engine(url)
    metadata.drop_all(engine)
    metadata.create_all(engine)



#!/usr/bin/env python3


def samples_by_date(from_date, to_date, page=0, size=100):
    filt_string = f"dt:update:from={from_date}until={to_date}"
    req = requests.get('https://www.ebi.ac.uk/biosamples/samples',
                       params = {"filter":filt_string, "size":size, "page": page})
    return req



class BaseType(object):
    def __init__(self, name, nullable):
        self.name = name
        self.nullable = nullable

    def __repr__(self):
        return "<{} class name:{}>".format(self.__type__, self.name)


################################################
#
#
#  Classes for simple entities:
#
#  Strings, Dates(timestamp), Boolean,
#  Long, Integer
#
#
################################################


class StringType(BaseType):
    __type__ = 'StringType'

    def __init__(self, name, nullable):
        super().__init__(name, nullable)

    def mapping(self):
        return ({
            "fields": {
                "keyword": {
                    "ignore_above": 256,
                    "type": "keyword"
                }
            },
            "type": "text"
        })


class DateType(BaseType):
    __type__ = 'DateType'

    def __init__(self, name, nullable):
        super().__init__(name, nullable)

    def mapping(self):
        return ({"type": "timestamp"})


class BooleanType(BaseType):
    __type__ = 'BooleanType'

    def __init__(self, name, nullable):
        super().__init__(name, nullable)

    def mapping(self):
        return ({"type": "boolean"})


class LongType(BaseType):
    __type__ = 'LongType'

    def __init__(self, name, nullable):
        super().__init__(name, nullable)

    def mapping(self):
        return ({"type": "long"})


class IntType(BaseType):
    __type__ = 'IntType'

    def __init__(self, name, nullable):
        super().__init__(name, nullable)

    def mapping(self):
        return ({"type": "integer"})


################################################
#
#
#  Classes for complex entities:
#
#    Objects
#    Nested (arrays)
#
#
################################################


class ObjectType(BaseType):
    # spark type: struct
    # ES type:    object
    __type__ = 'ObjectType'

    def __init__(self, name, nullable):
        super().__init__(name, nullable)
        self.fields = {}

    def mapping(self):
        return ({
            "properties":
            dict((k, self.fields[k].mapping()) for k in self.fields.keys())
        })


class NestedType(BaseType):
    # spark type: array
    # ES type:    nested
    __type__ = 'NestedType'

    def __init__(self, name, nullable):
        super().__init__(name, nullable)
        self.fields = {}

    def mapping(self):
        return ({
            "include_in_parent":
            True,
            "type":
            "nested",
            "properties":
            dict((k, self.fields[k].mapping()) for k in self.fields.keys())
        })


def walk_schema(schema, parent=None):
    """Walk a pyspark "schema.jsonValue" schema

    This function converts a pyspark schema into
    a set of classes that can then convert the schema
    to an elasticsearch mapping.

    Parameters
    ----------
    schema: a dict
        Typically, this would come from `df.schema.jsonValue()`. It could
        also be loaded from a string or file using the `json` library.
    parent: a container class, so an `ObjectType` or a `NestedType`
        If empty, will be initialized with an `ObjectType` with 
        name = "root".

    >>> s = walk_schema(df.schema.jsonValue(), ObjectType(name='root'))
    >>> s.mapping()
    """

    if (parent is None):
        parent = ObjectType(name='root')

    for field in schema['fields']:
        if (field['type'] == 'string'):
            parent.fields[field['name']] = StringType(
                name=field['name'], nullable=field['nullable'])
        if (field['type'] == 'integer'):
            parent.fields[field['name']] = IntType(name=field['name'],
                                                   nullable=field['nullable'])
        if (field['type'] == 'long'):
            parent.fields[field['name']] = LongType(name=field['name'],
                                                    nullable=field['nullable'])
        if (field['type'] == 'timestamp'):
            parent.fields[field['name']] = DateType(name=field['name'],
                                                    nullable=field['nullable'])
        if (field['type'] == 'boolean'):
            parent.fields[field['name']] = BooleanType(
                name=field['name'], nullable=field['nullable'])
        if (type(field['type']) == dict):
            n_field = field['type']
            if (n_field['type'] == 'struct'):
                newp = ObjectType(field['name'], field['nullable'])
                parent.fields[field['name']] = walk_schema(n_field,
                                                           parent=newp)

            if (n_field['type'] == 'array'):
                newp = NestedType(field['name'], field['nullable'])
                parent.fields[field['name']] = walk_schema(
                    n_field['elementType'], parent=newp)
    return parent



import requests
import gzip
import csv
import os
from databases import Database
import asyncio

from .db.models import (ontologies, parents, create_db,
                        synonyms, metadata, semantic_types)


database_url = 'postgresql://sdavis2@localhost/sdavis2'
database = Database(database_url)


def ontology_csv(ontology):
    custom_header = {'Accept-Encoding': 'gzip'}
    r = requests.get(
        f'http://data.bioontology.org/ontologies/{ontology}/'
        'download?apikey=8b5b7825-538d-40e0-9e9e-5ab9274a9aeb'
        '&download_format=csv',
        headers=custom_header,
        stream=True)
    # r.raw.decode_content = True
    try:
        f = gzip.GzipFile(fileobj=r.raw)
    except Exception:
        f = r.raw
    reader = csv.DictReader(line.decode('UTF-8') for line in f)
    reader.fieldnames = list(
        [n.lower().replace(' ', '_') for n in reader.fieldnames])
    for row in reader:
        row['synonyms'] = [st.strip('"') for st in row['synonyms'].split('|')]
        row['semantic_types'] = [st.strip('"')
                                 for st in row['semantic_types'].split('|')]
        row['parents'] = [st.strip('"') for st in row['parents'].split('|')]
        x = dict()
        for idx, key in enumerate(row.keys()):
            x['ontology'] = ontology
            if 0 <= idx < 8:
                if (row[key] == ''):
                    x[key] = None
                else:
                    x[key] = row[key]
                if (row[key] == ['']):
                    x[key] = []
        yield (x)


def get_all_ontology_short_ids():
    import yaml
    yaml_string = requests.get(
        'https://raw.githubusercontent.com/OBOFoundry/'
        'OBOFoundry.github.io/master/_config.yml'
    )
    val = yaml.load(yaml_string.content)
    x = []
    for onto in val['ontologies']:
        x.append(onto['id'].upper())
    return (x)


async def create_onto_table():
    await database.connect()
    # Create a table.
    await database.execute(query='DROP table ontologies CASCADE;')
    await database.execute(query='DROP table parents CASCADE;')
    await database.execute(query='DROP table synonyms CASCADE;')
    await database.execute(query='DROP table semantic_types CASCADE;')

    query = """
    CREATE TABLE ontologies (id SERIAL PRIMARY KEY,
        ontology varchar,
        class_id varchar,
        preferred_label varchar,
        definitions varchar,
        obsolete varchar,
        cui varchar)
    """
    await database.execute(query=query)
    query = """CREATE index ontologies_ontology_class on ontologies(ontology, class_id)"""
    await database.execute(query=query)
    query = """
    CREATE TABLE parents (id SERIAL PRIMARY KEY,
        ontology_id INTEGER references ontologies (id),
        parent varchar
    )"""
    await database.execute(query=query)
    query = """
    CREATE TABLE synonyms (id SERIAL PRIMARY KEY,
        ontology_id INTEGER references ontologies (id),
        synonym varchar
    )"""
    await database.execute(query=query)
    query = """
    CREATE TABLE semantic_types (id SERIAL PRIMARY KEY,
        ontology_id INTEGER references ontologies (id),
        semantic_type varchar
    )"""
    await database.execute(query=query)
    await database.disconnect()


async def main(vals):
    await database.connect()
    # Create a table.

# WITH data(firstname, lastname, adddetails, value) AS (
#    VALUES                              -- provide data here
#       ('fai55', 'shaggk', 'ss', 'ss2') -- see below
#     , ('fai56', 'XXaggk', 'xx', 'xx2') -- works for multiple input rows
#        --  more?
#    )
# , ins1 AS (
#    INSERT INTO sample (firstname, lastname)
#    SELECT firstname, lastname          -- DISTINCT? see below
#    FROM   data
#    -- ON     CONFLICT DO NOTHING       -- UNIQUE constraint? see below
#    RETURNING firstname, lastname, id AS sample_id
#    )
# , ins2 AS (
#    INSERT INTO sample1 (sample_id, adddetails)
#    SELECT ins1.sample_id, d.adddetails
#    FROM   data d
#    JOIN   ins1 USING (firstname, lastname)
#    RETURNING sample_id, user_id
#    )
# INSERT INTO sample2 (user_id, value)
# SELECT ins2.user_id, d.value
# FROM   data d
# JOIN   ins1 USING (firstname, lastname)
# JOIN   ins2 USING (sample_id);

    # Insert some data.
    query = """
    WITH data (ontology, class_id, preferred_label,
         synonyms, definitions, obsolete, cui, semantic_types, parents) as (
         VALUES (:ontology, :class_id, :preferred_label, CAST( :synonyms AS varchar[]),
         :definitions, :obsolete, :cui, CAST( :semantic_types AS varchar[]), 
         CAST( :parents as varchar[]))
    )
    , ins1 as (
        INSERT INTO ontologies (ontology, class_id, preferred_label,
         definitions, obsolete, cui)
         select ontology, class_id, preferred_label, definitions, obsolete, cui from data
         returning ontologies.id as ontology_id, class_id, ontology
    )
    , ins2 as (
        INSERT INTO parents (ontology_id, parent)
         select ontology_id, UNNEST(data.parents) 
        from ins1 join data USING(class_id,ontology))
    , ins3 as (
        INSERT INTO synonyms (ontology_id, synonym)
         select ontology_id, UNNEST(synonyms) 
        from ins1 join data USING(class_id,ontology))
    INSERT INTO semantic_types (ontology_id, semantic_type)
         select ontology_id, UNNEST(semantic_types) 
        from ins1 join data USING(class_id,ontology)
    """
    try:
        await database.execute_many(query=query, values=vals)
    except Exception as e:
        print(e)
    await database.disconnect()


if __name__ == '__main__':
    create_db(database_url)
    exit(0)
    asyncio.get_event_loop().run_until_complete(create_onto_table())
    for onto in sorted(get_all_ontology_short_ids()):
        print(onto)
        vals = []
        with open(onto + '.json', 'w') as outfile:
            try:
                vals = ontology_csv(onto)
                asyncio.get_event_loop().run_until_complete(main(vals))
            except KeyboardInterrupt:
                exit(0)
            except Exception as e:
                os.unlink(onto + '.json')
                print(f'error with {onto}')
                print(e)
                continue
        



# coding: utf-8
from sqlalchemy import BigInteger, Column, DateTime, ForeignKey, Integer, Table, Text, text, UniqueConstraint
from sqlalchemy.dialects.postgresql import JSONB, TSVECTOR
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base
import sqlalchemy.dialects.postgresql as postgresql

Base = declarative_base()
metadata = Base.metadata

etl_schema = 'etl'
public_schema = 'public2'
staging_schema = 'staging'


class IdentifiersMixin(object):
    identifier = Column(Text)
    namespace = Column(Text)


class XrefsMixin(object):
    db = Column(Text)
    identifier = Column(Text)
    uuid = Column(Text)


class AttributesMixin(object):
    tag = Column(Text)
    value = Column(Text)


t_biosample_jsonb = Table('biosample_jsonb',
                          metadata,
                          Column('doc', JSONB(astext_type=Text())),
                          schema=etl_schema)

t_biosample_test = Table('biosample_test',
                         metadata,
                         Column('id', Integer),
                         Column('gsm', Text),
                         Column('dbgap', Text),
                         Column('model', Text),
                         Column('title', Text),
                         Column('access', Text),
                         Column('id_recs', JSONB(astext_type=Text())),
                         Column('ids', JSONB(astext_type=Text())),
                         Column('taxon_id', Integer),
                         Column('accession', Text),
                         Column('attributes', JSONB(astext_type=Text())),
                         Column('sra_sample', Text),
                         Column('description', Text),
                         Column('last_update', DateTime),
                         Column('publication_date', DateTime),
                         Column('submission_date', DateTime),
                         Column('is_reference', Text),
                         Column('taxonomy_name', Text),
                         Column('textsearchable_index_col',
                                TSVECTOR,
                                index=True),
                         schema=etl_schema)

t_experiment_jsonb = Table('experiment_jsonb',
                           metadata,
                           Column('doc', JSONB(astext_type=Text())),
                           schema=etl_schema)

t_fileinfo_addons = Table('fileinfo_addons',
                          metadata,
                          Column('accession', Text, index=True),
                          Column('file_size', BigInteger),
                          Column('file_md5', Text),
                          Column('file_date', DateTime),
                          Column('filename', Text),
                          Column('url', Text),
                          schema=etl_schema)

t_fileinfo_runs = Table('fileinfo_runs',
                        metadata,
                        Column('accession', Text, index=True),
                        Column('file_size', BigInteger),
                        Column('file_md5', Text),
                        Column('file_date', DateTime),
                        schema=etl_schema)

t_geo_jsonb = Table(
    'geo_jsonb',
    metadata,
    Column('doc', JSONB(astext_type=Text())),
    Column('update_time',
           DateTime,
           server_default=text(
               "'2019-03-20 12:55:47.158602'::timestamp without time zone")),
    schema=etl_schema)

t_run_jsonb = Table('run_jsonb',
                    metadata,
                    Column('doc', JSONB(astext_type=Text())),
                    schema=etl_schema)

t_sample_jsonb = Table('sample_jsonb',
                       metadata,
                       Column('doc', JSONB(astext_type=Text())),
                       schema=etl_schema)


class SraAccession(Base):
    __tablename__ = 'sra_accession'
    __table_args__ = {'schema': etl_schema}

    accession = Column(Text, primary_key=True, index=True)
    submission = Column(Text)
    status = Column(Text)
    updated = Column(DateTime)
    published = Column(DateTime)
    received = Column(DateTime)
    type = Column(Text)
    center = Column(Text)
    visibility = Column(Text)
    alias = Column(Text)
    experiment = Column(Text)
    sample = Column(Text)
    study = Column(Text)
    loaded = Column(Text)
    spots = Column(BigInteger)
    bases = Column(BigInteger)
    md5sum = Column(Text)
    bio_sample = Column(Text)
    bio_project = Column(Text)
    replaced_by = Column(Text)


class SraStudy(Base):
    __tablename__ = 'sra_study'
    __table_args__ = {'schema': public_schema}

    bioproject = Column(Text)
    gse = Column(Text)
    abstract = Column(Text)
    accession = Column(Text, primary_key=True)
    alias = Column(Text)
    attributes = Column(JSONB(astext_type=Text()))
    broker_name = Column(Text)
    center_name = Column(Text)
    description = Column(Text)
    identifiers = Column(JSONB(astext_type=Text()))
    study_type = Column(Text)
    title = Column(Text)
    xrefs = Column(JSONB(astext_type=Text()))
    status = Column(Text)
    updated = Column(DateTime)
    published = Column(DateTime)
    received = Column(DateTime)
    visibility = Column(Text)
    bio_project = Column(Text)
    replaced_by = Column(Text)


class SraSample(Base):
    __tablename__ = 'sra_sample'
    __table_args__ = {'schema': public_schema}

    accession = Column(Text, primary_key=True)
    alias = Column(Text)
    attributes = Column(JSONB(astext_type=Text()))
    bio_sample = Column(Text)
    broker_name = Column(Text)
    center_name = Column(Text)
    description = Column(Text)
    gsm = Column(Text)
    identifiers = Column(JSONB(astext_type=Text()))
    organism = Column(Text)
    title = Column(Text)
    taxon_id = Column(Integer)
    xrefs = Column(JSONB(astext_type=Text()))
    status = Column(Text)
    updated = Column(DateTime)
    published = Column(DateTime)
    received = Column(DateTime)
    visibility = Column(Text)
    replaced_by = Column(Text)
    study_accession = Column(ForeignKey(public_schema +
                                        '.sra_study.accession'))

    sra_study = relationship('SraStudy')


t_study_jsonb = Table('study_jsonb',
                      metadata,
                      Column('doc', JSONB(astext_type=Text())),
                      schema=etl_schema)


class SraExperiment(Base):
    __tablename__ = 'sra_experiment'
    __table_args__ = {'schema': public_schema}

    accession = Column(Text, primary_key=True)
    alias = Column(Text)
    attributes = Column(JSONB(astext_type=Text()))
    broker_name = Column(Text)
    center_name = Column(Text)
    description = Column(Text)
    design = Column(Text)
    identifiers = Column(JSONB(astext_type=Text()))
    instrument_model = Column(Text)
    library_construction_protocol = Column(Text)
    library_layout_length = Column(Text)
    library_layout_orientation = Column(Text)
    library_layout_sdev = Column(Text)
    library_name = Column(Text)
    library_strategy = Column(ForeignKey(public_schema +
                                         '.sra_library_strategy.value'),
                              index=True)
    library_source = Column(ForeignKey(public_schema +
                                       '.sra_library_source.value'),
                            index=True)
    library_layout = Column(ForeignKey(public_schema +
                                       '.sra_library_layout.value'),
                            index=True)
    library_selection = Column(ForeignKey(public_schema +
                                          '.sra_library_selection.value'),
                               index=True)
    platform = Column(ForeignKey(public_schema + '.sra_platform.value'),
                      index=True)
    sample_accession = Column(
        ForeignKey(public_schema + '.sra_sample.accession'))
    study_accession = Column(ForeignKey(public_schema +
                                        '.sra_study.accession'))
    title = Column(Text)
    xrefs = Column(Text)
    status = Column(Text)
    updated = Column(DateTime)
    published = Column(DateTime)
    received = Column(DateTime)
    visibility = Column(Text)
    replaced_by = Column(Text)

    sra_sample = relationship('SraSample')
    sra_study = relationship('SraStudy')


class SraRun(Base):
    __tablename__ = 'sra_run'
    __table_args__ = {'schema': public_schema}

    accession = Column(Text, primary_key=True)
    broker_name = Column(Text)
    identifiers = Column(JSONB(astext_type=Text()))
    run_center = Column(Text)
    center_name = Column(Text)
    run_date = Column(DateTime)
    attributes = Column(JSONB(astext_type=Text()))
    nreads = Column(Integer)
    alias = Column(Text)
    spot_length = Column(Integer)
    experiment_accession = Column(
        ForeignKey(public_schema + '.sra_experiment.accession'))
    study_accession = Column(ForeignKey(public_schema +
                                        '.sra_study.accession'))
    sample_accession = Column(
        ForeignKey(public_schema + '.sra_sample.accession'))
    reads = Column(JSONB(astext_type=Text()))
    status = Column(Text)
    updated = Column(DateTime)
    published = Column(DateTime)
    received = Column(DateTime)
    visibility = Column(Text)
    bio_project = Column(Text)
    replaced_by = Column(Text)
    loaded = Column(Text)
    spots = Column(BigInteger)
    bases = Column(BigInteger)

    sra_experiment = relationship('SraExperiment')
    sra_study = relationship('SraStudy')
    sra_sample = relationship('SraSample')


class SraLibraryStrategy(Base):
    __tablename__ = 'sra_library_strategy'
    __table_args__ = {'schema': public_schema}

    value = Column(Text, primary_key=True)


class SraLibrarySource(Base):
    __tablename__ = 'sra_library_source'
    __table_args__ = {'schema': public_schema}

    value = Column(Text, primary_key=True)


class SraLibraryLayout(Base):
    __tablename__ = 'sra_library_layout'
    __table_args__ = {'schema': public_schema}

    value = Column(Text, primary_key=True)


class SraLibrarySelection(Base):
    __tablename__ = 'sra_library_selection'
    __table_args__ = {'schema': public_schema}

    value = Column(Text, primary_key=True)


class SraPlatform(Base):
    __tablename__ = 'sra_platform'
    __table_args__ = {'schema': public_schema}

    value = Column(Text, primary_key=True)


#########################
# GEO models below here #
#########################

geo_gse_gsm = Table('geo_gse_gsm', Base.metadata,
                    Column('gse', Text, ForeignKey('geo_gse.accession')),
                    Column('gsm', Text, ForeignKey('geo_gsm.accession')))

geo_gse_gpl = Table('geo_gse_gpl', Base.metadata,
                    Column('gse', Text, ForeignKey('geo_gse.accession')),
                    Column('gpl', Text, ForeignKey('geo_gpl.accession')))


class GEOSeries(Base):
    __tablename__ = 'geo_gse'
    __table_args__ = {'schema': public_schema}

    accession = Column(Text, primary_key=True)
    title = Column(Text)
    status = Column(Text)
    summary = Column(Text)
    description = Column(Text)
    overall_design = Column(Text)
    submission_date = Column(DateTime)
    last_update_date = Column(DateTime)
    data_processing = Column(Text)
    contributors = Column(postgresql.ARRAY(Text))
    gpls = relationship("GEOPlatform",
                        secondary=geo_gse_gpl,
                        back_populates="gses")
    gsms = relationship("GEOSample",
                        secondary=geo_gse_gsm,
                        back_populates="gses")


class GEOPlatform(Base):
    __tablename__ = 'geo_gpl'
    __table_args__ = {'schema': public_schema}

    accession = Column(Text, primary_key=True)
    title = Column(Text)
    status = Column(Text)
    summary = Column(Text)
    description = Column(Text)
    submission_date = Column(DateTime)
    last_update_date = Column(DateTime)
    technology = Column(Text)
    distribution = Column(Text)
    data_row_count = Column(Integer)
    contributors = Column(postgresql.ARRAY(Text))
    gses = relationship("GEOSeries",
                        secondary=geo_gse_gpl,
                        back_populates="gpls")
    gsms = relationship("GEOSample",
                        secondary=geo_gse_gsm,
                        back_populates="gpls")


class GEOContact(Base):
    __tablename__ = 'geo_contact'
    __table_args__ = {'schema': public_schema}

    email = Column(Text, primary_key=True)
    name = Column(Text)
    phone = Column(Text)
    fax = Column(Text)
    city = Column(Text)
    address = Column(Text)
    country = Column(Text)
    web_link = Column(Text)
    institute = Column(Text)
    postal_code = Column(Text)


biosample_to_attributes = Table(
    'biosample_to_attributes',
    Base.metadata,
    Column('id', Integer, primary_key=True),
    Column('biosample_id', ForeignKey('biosample.id')),
    Column('biosample_attribute_id', ForeignKey('biosample_attribute.id')),
)


class BiosampleModel(Base):
    __tablename__ = "biosample_model"

    id = Column(Integer, primary_key=True)
    name = Column(Text, index=True)


class Biosample(Base):
    __tablename__ = "biosample"

    id = Column(BigInteger, primary_key=True)
    submission_date = Column(DateTime, index=True, nullable=False)
    last_update = Column(DateTime, index=True, nullable=False)
    publication_date = Column(DateTime, index=True, nullable=False)
    access = Column(Text,
                    index=True,
                    comment="access control, public or controlled")
    # srs = Column(ForeignKey('sra_sample.accession'), index=True)
    # gsm = Column(ForeignKey('geo_gsm.accession'), index=True)
    dbgap = Column(Text, index=True, comment="dbGaP identifier")
    title = Column(Text)
    description = Column(Text)
    taxonomy_name = Column(Text)
    taxon_id = Column(Integer, index=True, comment="NCBI Taxonomy id")
    model = Column(ForeignKey("biosample_model.id"), index=True)
    attributes = relationship("BiosampleAttribute",
                              secondary=biosample_to_attributes)


class BiosampleAttribute(Base):
    __tablename__ = "biosample_attribute"

    id = Column(Integer, primary_key=True)
    name = Column(Text, index=True)
    harmonized_name = Column(Text, index=True)
    display_name = Column(Text, index=True)
    value = Column(Text, index=True)
    UniqueConstraint('name', 'harmonized_name', 'value')
    biosamples = relationship("Biosample", secondary=biosample_to_attributes)


class BiosampleIdentifier(Base):
    __tablename__ = "biosample_identifier"

    id = Column(Integer, primary_key=True)
    biosample = Column(ForeignKey('biosample.id'), index=True)
    db = Column(Text, index=True)
    label = Column(Text, index=True)
    identifier = Column(Text, index=True)



"""Usage

"""

from Bio import Entrez
import logging
import collections
import re
import datetime
import urllib
from xml import etree
import xml
from io import StringIO
import requests

import omicidx.geo.pydantic_models as pydantic_models

logging.basicConfig(level=logging.INFO)

try:
    from urllib.error import HTTPError  # for Python 3
except ImportError:
    from urllib2 import HTTPError  # for Python 2

ETYP = ['GSE', 'GSM', 'GPL', 'GDS']


def get_entrez_instance(email='user@example.com'):
    """Return a Bio::Entrez object
    
    # Arguments
    email (str): the email to be used with the Entrez instance
    
    Returns
    =======
    A Bio::Entrez instance
    """
    ret = Entrez
    ret.email = email
    return ret


def get_geo_accessions(etyp='GSE',
                       batch_size=1000,
                       add_term=None,
                       email="user@example.com"):
    """Get GEO accessions by etyp

    Useful for getting all the ETYP accessions for
    later bulk processing

    Parameters
    ----------
    etyp: str
        One of GSE, GPL, GSM, GDS
    batch_size: int 
        the number of accessions to return in one batch. 
        Transparent to the user, as this returns an iterator.
    add_term: str
        Add a search term for the query. Useful to limit
        by date or search for specific text. For example, 
        to limit by date: '2007/01/01:2007/03/01[PDAT]'
    email: str
        user email (not important)

    Return
    ------
    an iterator of accessions, each as a string
    """

    entrez = get_entrez_instance(email)
    if (etyp not in ETYP):
        raise Exception("etype {} not one of the accepted etyps {}".format(
            etyp, str(ETYP)))
    handle = entrez.esearch(db='gds',
                            term=etyp + '[ETYP]' + str(add_term),
                            usehistory='y')
    search_results = entrez.read(handle)
    webenv = search_results["WebEnv"]
    query_key = search_results["QueryKey"]
    count = int(search_results['Count'])
    data = []
    logging.info('found {} records for {} database'.format(count, etyp))
    from IPython.core.debugger import set_trace
    n = 0
    for start in range(0, count, batch_size):
        end = min(count, start + batch_size)
        attempt = 0
        fetch_handle = None
        while attempt < 10:
            attempt += 1
            try:
                fetch_handle = entrez.esummary(db="gds",
                                               rettype="acc",
                                               retmode="xml",
                                               retstart=start,
                                               retmax=batch_size,
                                               webenv=webenv,
                                               query_key=query_key)
                for g in entrez.read(fetch_handle):
                    n += 1
                    yield (g['Accession'])

                break
            except (HTTPError, RuntimeError) as err:
                import time
                if (isinstance(err, HTTPError)):
                    logging.error("Received error from server %s" % err)
                    logging.error("Attempt %i of 10" % attempt)
                    time.sleep(1 * attempt * attempt)
                else:
                    logging.error("RuntimeError received: {}".format(err))
                    logging.error("Attempt %i of 10" % attempt)
                    time.sleep(1)
            else:
                raise


def get_geo_accession_xml(accession, targ='all', view='brief'):
    """Open a connection to get the GEO XML for an accession

    Parameters
    ==========
    accession: str the GEO accesssion
    targ: str what to fetch. One of "all", "series", "platform", 
        "samples", "self"
    view: str amount to return.

    Returns
    =======
    A file-like object for reading or readlines
    
    >>> handle = get_geo_accession_xml('GSE2553')
    >>> handle.readlines()
    """
    url = (
        "https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?targ={}&acc={}&form=xml&view={}"
        .format(targ, accession, view))
    attempt = 0
    while attempt < 10:
        attempt += 1
        try:
            return urllib.request.urlopen(url)
            break
        except Exception as err:
            print("Received error from server %s" % err)
            print("Attempt %i of 3" % attempt)
            import time
            time.sleep(1 * attempt)


def get_geo_accession_soft(accession, targ='all', view="brief"):
    """Open a connection to get the GEO SOFT for an accession

    Parameters
    ==========
    accession: str the GEO accesssion
    targ: str what to fetch. One of "all", "series", "platform", 
        "samples", "self"
    view: str amount to return.

    Returns
    =======
    A file-like object for reading or readlines
    
    >>> handle = get_geo_accession_soft('GSE2553')
    >>> handle.readlines()
    """
    logging.info(f'Accessing accession {accession} in SOFT format')
    url = (
        "https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?targ={}&acc={}&form=text&view={}"
        .format(targ, accession, view))
    attempt = 0
    while attempt < 10:
        attempt += 1
        try:
            resp = requests.get(url)
            return StringIO(resp.text)
            break
        except Exception as err:
            print("Received error from server %s" % err)
            print("Attempt %i of 10" % attempt)
            import time
            time.sleep(1 * attempt)


def _split_on_first_equal(line, val='='):
    """Split a line based on first occurrence of " ="

    Parameters
    ----------
    line: str
        string to split
    split: str
        split based on this. Can be multiple characters.
    >>> _split_on_first_equal("this = abc = 1")
    ('this', 'abc = 1')
    """
    l = line.split(" {} ".format(val))
    if (l[0].endswith(" {}".format(val))):
        l[0] = l[0].replace(' {}'.format(val), '')
    return (l[0], " {} ".format(val).join(l[1:]))


def get_geo_entities(txt):
    """Get the text associated with each entity from a block of text

    Parameters
    ----------
    txt: list(str)
    
    Returns
    -------
    A dict of entities keyed by accession and values a list of txt lines
    """
    logging.info('Parsing out geo entities')
    entities = {}
    accession = None
    entity = []
    for line in txt:
        line = line.strip()
        if (line.startswith('^')):
            accession = _split_on_first_equal(line)[1]
            entities[accession] = entity
            entity = []
        entity.append(line)
    entities[accession] = entity
    logging.info(f'found {len(entities.keys())}')
    return entities


#####################################
# Utility functions for GSE parsing #
#####################################


def get_subseries_from_relations(relation_list):
    ret = []
    for i in relation_list:
        if (i.startswith("SuperSeries of:")):
            ret.append(i.replace('SuperSeries of: ', ''))
    return ret


def get_bioprojects_from_relations(relation_list):
    ret = []
    for i in relation_list:
        if (i.startswith(
                "BioProject: https://www.ncbi.nlm.nih.gov/bioproject/")):
            ret.append(
                i.replace(
                    'BioProject: https://www.ncbi.nlm.nih.gov/bioproject/',
                    ''))
    return ret


def get_SRA_from_relations(relation_list):
    ret = []
    for i in relation_list:
        if (i.startswith("SRA: https://www.ncbi.nlm.nih.gov/sra?term=")):
            ret.append(
                i.replace('SRA: https://www.ncbi.nlm.nih.gov/sra?term=', ''))
    return ret


def get_biosample_from_relations(relation_list):
    ret = []
    for i in relation_list:
        if (i.startswith("BioSample: https://www.ncbi.nlm.nih.gov/biosample/")
            ):
            ret.append(
                i.replace('BioSample: https://www.ncbi.nlm.nih.gov/biosample/',
                          ''))
    return ret


def get_channel_characteristics(d, ch):
    ch_items = list([k for k in d.keys() if k.endswith('ch{}'.format(ch))])
    characteristics = []
    ret = {}
    for k in ch_items:
        if (k.startswith('characteristics')):
            for characteristic in d[k]:
                characteristics.append(tuple(characteristic.split(': ')))
            continue
        newk = k.replace("_ch{}".format(ch), "")
        ret[k.replace("_ch{}".format(ch), "")] = "\n".join(d[k])
        if (newk == 'taxid'):
            ret[newk] = list([int(x) for x in ret[newk].split('\n')])
    char = []
    for v in characteristics:
        if (len(v) == 1):
            tag = v[0]
            val = None
        elif (len(v) == 2):
            tag = v[0]
            val = v[1]
        elif (len(v) > 2):
            tag = v[0]
            val = ":".join(v[1:])
        char.append({"tag": tag, "value": val})
    ret['characteristics'] = char
    return ret


def _split_geo_name(v):
    """split name of form first,middle,last into dict"""
    return dict(zip('first middle last'.split(), v.split(',')))


def _create_contact_from_parsed(d):
    contact_dict = {}
    for k, v in d.items():
        if (k.startswith('contact')):
            v = v[0]
            if (v == ''):
                v = None
            if (k == 'contact_name'):
                v = _split_geo_name(v)
            if (k == 'contact_web_link'):
                if (not v.startswith('http')):
                    v = 'http://' + v
            contact_dict[k.replace('contact_', '')] = v
    return contact_dict


def _split_contributor_names(contribs):
    if (len(contribs) == 0):
        return []
    return list([_split_geo_name(v) for v in contribs])


###############################################
# Parse a single entity of SOFT format,       #
# generically. Then pass along to             #
# GSE, GSM, or GPL specific parser and return #
# appropriate class.                          #
###############################################
def _parse_single_entity_soft(entity_txt):
    # Deal with common stuff first:
    tups = [_split_on_first_equal(line) for line in entity_txt]
    accession = tups[0][1]
    entity_type = tups[0][0].replace('^', '')
    tups = [(re.sub('![^_]+_', '', tup[0]), tup[1]) for tup in tups]
    d = collections.defaultdict(list)
    for tup in tups[1:]:
        d[tup[0]].append(tup[1])
    d2 = dict((k, v) for k, v in d.items())
    # int fields
    INT_FIELDS = ['pubmed_id', 'sample_taxid', 'platform_taxid']
    for i in INT_FIELDS:
        if (i in d2):
            try:
                d2[i] = list(map(lambda x: int(x), d2[i]))
            except:
                d2[i] = []
    # change "geo_accesion = [...]" to "accession = ..."
    try:
        d2['accession'] = d2['accession'][0]
    except KeyError:
        d2['accession'] = d2['geo_accession'][0]
        del (d2['geo_accession'])
    if ('description =' in d2):
        del (d2['description ='])
    for k in [
            'status',
            'description',
            'data_processing',
            'title',
            'summary',  # GSE only?
            'submission_date',  # to fix later
            'last_update_date',  # to fix later
            'overall_design'
    ]:  #GSE only ]:
        try:
            d2[k] = "\n".join(d2[k])
        except KeyError:
            d2[k] = None
    # contact details as object
    contact = _create_contact_from_parsed(d2)
    dkeys = d2.keys()
    for k in list(dkeys):
        if k.startswith('contact'):
            del (d2[k])
    d2['contact'] = contact
    if (entity_type == 'SERIES'):
        return (_parse_single_gse_soft(d2))
    if (entity_type == 'SAMPLE'):
        return (_parse_single_gsm_soft(d2))
    if (entity_type == 'PLATFORM'):
        return (_parse_single_gpl_soft(d2))


###############################################################################
#        Update date fields of format 'month day year' to date type.          #
###############################################################################
def _fix_date_fields(d):
    for datefield in ['submission_date', 'last_update_date']:
        if datefield in d:
            d[datefield] = datetime.datetime.strptime(d[datefield], '%b %d %Y')
    return d


#######################################
# Parse the GSE entity in SOFT format #
#######################################
def _parse_single_gse_soft(d2):
    try:
        d2['subseries'] = get_subseries_from_relations(d2['relation'])
        d2['bioprojects'] = get_bioprojects_from_relations(d2['relation'])
        d2['sra_studies'] = get_SRA_from_relations(d2['relation'])
        # GEO sometimes references SRS and SRR from series--filter those out
        d2['sra_studies'] = list(
            filter(lambda a: a.find('P') > 0, d2['sra_studies']))
    except KeyError:
        d2['subseries'] = []
        d2['bioprojects'] = []
        d2['sra_studies'] = []
    supp_files = []
    for k in list(d2.keys()):
        if (k.startswith('supplementa')):
            supp_files += d2[k]
            del d2[k]
    d2['supplemental_files'] = supp_files
    if (len(supp_files) > 0):
        if (supp_files[0] == 'NONE'):
            supp_files = []
    d2['_entity'] = 'GSE'
    if ('contributor' in d2):
        d2['contributor'] = _split_contributor_names(d2['contributor'])
    d2 = _fix_date_fields(d2)
    return pydantic_models.GEOSeries(**d2)


def _create_gsm_channel_data(d):
    ch_count = int(d['channel_count'])
    ch_recs = []
    for i in range(0, ch_count):
        # this GEOChannel just encapsulates the parsing
        # of channel records.
        # TODO: refactor to parse directly to dict
        ch_recs.append(get_channel_characteristics(d, i + 1))
    return ch_recs


# Search for fields like _ch1 and _ch2 ....
import re
r1 = re.compile(r'_ch\d+$')


#######################################
# Parse the GSM entity in SOFT format #
#######################################
def _parse_single_gsm_soft(d2):
    # to singular:
    for k in [
            'platform_id', 'channel_count', 'library_strategy',
            'library_source', 'library_selection', 'instrument_model',
            'data_row_count', 'anchor', 'tag_length', 'type', 'tag_count'
    ]:
        try:
            d2[k] = "\n".join(d2[k])
        except KeyError:
            d2[k] = None
    try:
        d2['biosample'] = get_biosample_from_relations(d2['relation'])[0]
        d2['sra_experiment'] = get_SRA_from_relations(d2['relation'])[0]
    except:
        d2['biosample'] = None
        d2['sra_experiment'] = None
    if ('contributor' in d2):
        d2['contributor'] = _split_contributor_names(d2['contributor'])
    d2['channels'] = _create_gsm_channel_data(d2)
    supp_files = []
    for k in list(d2.keys()):
        if r1.search(
                k
        ):  # have to use search here because endswith does not take regex
            del d2[k]
        if (k.startswith('supplementa')):
            supp_files += d2[k]
            del d2[k]
    d2['supplemental_files'] = supp_files
    if (supp_files[0] == 'NONE'):
        supp_files = []
    d2['_entity'] = 'GSM'
    d2 = _fix_date_fields(d2)
    # Collapse protocol fields into single string
    for k in filter(lambda k: k.endswith('protocol'), d2.keys()):
        if isinstance(d2[k], list):
            d2[k] = ' '.join(d2[k])
    if ('contributor' in d2):
        d2['contributor'] = _split_contributor_names(d2['contributor'])
    return pydantic_models.GEOSample(**d2)


#######################################
# Parse the GPL entity in SOFT format #
#######################################
def _parse_single_gpl_soft(d2):
    for k in [
            'technology', 'distribution', 'organism', 'taxid', 'data_row_count'
    ]:
        try:
            d2[k] = "\n".join(d2[k])
        except KeyError:
            d2[k] = None
    d2['_entity'] = 'GPL'
    d2 = _fix_date_fields(d2)
    if ('manufacture_protocol' in d2):
        d2['manufacture_protocol'] = "\n".join(d2['manufacture_protocol'])
    if ('contributor' in d2):
        d2['contributor'] = _split_contributor_names(d2['contributor'])
    return pydantic_models.GEOPlatform(**d2)


###############################################################################
#                   This is the main entrypoint for parsing.                  #
###############################################################################
def geo_entity_iterator(geo: str, targ: str = 'self', view: str = 'brief'):
    """Returns an iterator of GEO entities

    Given a GEO accession (typically a GSE,
    will return an iterator of the GEO entities
    associated with the record, including all
    GSMs, GPLs, and the GSE record itself

    Parameters
    ----------
    fh: Iterable[str]
       Anything that can iterate over lines of text
       Could be a list of text lines, a file handle, or
       an open url.
    
    Yields
    ------
    Iterator of GEO entities as pydantic models


    >>> for i in geo_soft_entity_iterator('GSE2553'):
    ...     print(i)
    """
    entity = []
    accession = None
    in_table = False
    fh = get_geo_accession_soft(geo, targ=targ, view=view)
    for line in fh:
        try:
            if (isinstance(line, bytes)):
                line = line.decode()
        except:
            pass
        line = line.strip()

        # ignore table details for now

        if (line.endswith('table_begin')):
            in_table = True
        if (line.endswith('table_end')):
            in_table = False
            continue
        if (in_table):
            continue

        # ignore comments

        if (line.startswith('#')):
            continue

        if (line.startswith('^')):
            if (accession is not None):
                yield (_parse_single_entity_soft(entity))
            accession = _split_on_first_equal(line)[1]
            entity = []

        entity.append(line)
    yield (_parse_single_entity_soft(entity))


def gse_to_json(gse):
    """Return json version of GSE record as text
    
    Parameters
    ----------
    gse : str
        The GSE accession to convert
    """
    res = []
    for z in geo_soft_entity_iterator(get_geo_accession_soft(gse)):
        res.append(z.json())
    return ("\n".join(res))


def bulk_gse_to_json(term, outfile='/dev/stdout'):
    """Bulk process GSEs to json

    This function takes a term (see below) and does a search. Results
    from the search of GSE records are returned as json lines to the
    output file in `outfile`.

    Parameters
    ----------
    term : str
        A search string like "2019/01/28:2019/02/28[PDAT]" as an example.
        Passed to entrez search.
    outfile: str
        The output file to which to write
    """
    for gse in get_geo_accessions(add_term=term):
        if (outfile is not None):
            logging.info(f'writing accession {gse}')
            res = gse_to_json(gse)
            if (res is not None):
                outfile.write(res + "\n")
            else:
                logging.warning(f'Accession {gse} not found, it appears.')


def gse_accessions(term, outfile='/dev/stdout'):
    """Output list of gse accessions based on term search

    Writes to a file.

    Parameters
    ----------
    term : str
        A search string like "2019/01/28:2019/02/28[PDAT]" as an example.
        Passed to entrez search.
    outfile: str
        The output file to which to write

    """

    with open(outfile, 'w') as ofile:
        n = 0
        for gse in get_geo_accessions(add_term=term, batch_size=1000):
            n += 1
            if (n % 5000 == 0):
                logging.info(f'Processed {n} records')
            ofile.write(gse + "\n")
        logging.info(f'Processed {n} records')



"""Biosample parser

Implemented as an iterator


>>> import omicidx.biosample as b
>>> for bios in b.BioSampleParser('biosample_set.xml.gz'):
>>>     print(bios.as_json())
"""

import os.path
import xml.etree.ElementTree as ET
import gzip
import json
import logging
import click
import subprocess


class BioSample(dict):
    """BioSample class"""
    def __init__(self):
        self.update()

    def as_json(self, indent=None):
        return (json.dumps(self))

    def __str__(self):
        return ('{0}'.format(self['title']))

    def __repr__(self):
        return ('BioSample <id={0} title={1}>'.format(self['id'],
                                                      self['title']))


class BioSampleParser(object):
    """Parse a BioSample xml file.

    Implemented as an iterator"""
    def __init__(self, fname):
        """Init

        Handles .gz as well as plain text"""
        self.fhandle = None
        if (fname.endswith('gz')):
            self.fhandle = gzip.open(fname)
        else:
            self.fhandle = open(fname)
        self.context = ET.iterparse(self.fhandle, events=("start", "end"))
        event, self.root = next(self.context)

    def __iter__(self):
        return (self)

    def __next__(self):
        for event, elem in self.context:
            if event == "end" and elem.tag == "BioSample":
                bios = BioSample()
                bios['is_reference'] = None
                for k, v in elem.items():
                    bios[k] = v
                bios['id_recs'] = []
                bios['ids'] = []
                bios['sra_sample'] = None
                bios['dbgap'] = None
                bios['gsm'] = None
                for id in elem.iterfind('.//Id'):
                    idrec = {
                        'db': id.get('db'),
                        'label': id.get('db_label'),
                        'id': id.text
                    }
                    bios['ids'].append(idrec['id'])
                    bios['id_recs'].append(idrec)
                    # add xref fields for SRA, dbGaP, and GEO

                    if (id.get('db') == 'SRA'):
                        bios['sra_sample'] = id.text
                    if (id.get('db') == 'dbGaP'):
                        bios['dbgap'] = id.text
                    if (id.get('db') == 'GEO'):
                        bios['gsm'] = id.text
                bios['title'] = elem.findtext('.//Description/Title')
                bios['description'] = elem.findtext(
                    './/Description/Comment/Paragraph')
                organism = elem.find('.//Organism')
                bios['taxonomy_name'] = organism.get('taxonomy_name')
                bios['taxon_id'] = int(organism.get('taxonomy_id'))
                bios['attribute_recs'] = []
                bios['attributes'] = []
                for attribute in elem.findall('.//Attribute'):
                    attrec = attribute.attrib
                    attrec['value'] = attribute.text
                    bios['attribute_recs'].append(attrec)
                    try:
                        bios['attributes'].append(attrec['harmonized_name'])
                    except:
                        bios['attributes'].append(attrec['attribute_name'])

                bios['model'] = elem.findtext('.//Model')
                #print(json.dumps(bios))
                #res = es.index(index="bioes", doc_type='biosample', id=bios['id'], body=bios)
                elem.clear()
                return bios


from pydantic import BaseModel
from typing import List


class BioProject(BaseModel):
    name: str
    title: str
    description: str = None
    pubs: List[dict]


class BioProjectParser(object):
    """Parse a BioProject xml file.

    Implemented as an iterator"""
    def __init__(self, fname):
        """Init

        Handles .gz as well as plain text"""
        self.fhandle = None
        if (fname.endswith('gz')):
            self.fhandle = gzip.open(fname)
        else:
            self.fhandle = open(fname)
        self.context = ET.iterparse(self.fhandle, events=("start", "end"))
        event, self.root = next(self.context)

    def __iter__(self):
        return (self)

    def __next__(self):
        for event, elem in self.context:
            if event == "end" and elem.tag == "Package":
                projtop = elem.find('./Project')
                d2 = {}
                d2['title'] = projtop.findtext('./Project/ProjectDescr/Title')
                d2['description'] = projtop.findtext(
                    './Project/ProjectDescr/Description')
                d2['name'] = projtop.findtext('./Project/ProjectDescr/Name')
                pubs = []
                for pub in projtop.findall(".//Publication"):
                    pubs.append({
                        "pubdate":
                        pub.attrib.get('date',None),
                        "id":
                        pub.attrib.get('id',None),
                        "db":
                        pub.findtext('./DbType').replace('^e', '', 1)
                    })
                d2['publications'] = pubs
                ext_links = []
                for link in projtop.findall(".//ExternalLink"):
                    ext_links.append({
                        "category": link.attrib['category'],
                        "label": link.attrib['label'],
                        "url": link.findtext('./URL')
                    })
                d2['external_links'] = ext_links
                return d2

                # bios = BioProject()
                # bios['is_reference']=None
                # for k,v in elem.items():
                #     bios[k] = v
                # bios['id_recs'] = []
                # bios['ids'] = []
                # bios['sra_sample']=None
                # bios['dbgap']=None
                # bios['gsm']=None
                # for id in elem.iterfind('.//Id'):
                #     idrec = {'db': id.get('db'), 'label': id.get('db_label'), 'id':id.text}
                #     bios['ids'].append(idrec['id'])
                #     bios['id_recs'].append(idrec)
                #     # add xref fields for SRA, dbGaP, and GEO

                #     if(id.get('db')=='SRA'):
                #         bios['sra_sample']=id.text
                #     if(id.get('db')=='dbGaP'):
                #         bios['dbgap']=id.text
                #     if(id.get('db')=='GEO'):
                #         bios['gsm']=id.text
                # bios['title'] = elem.findtext('.//Description/Title')
                # bios['description'] = elem.findtext('.//Description/Comment/Paragraph')
                # organism = elem.find('.//Organism')
                # bios['taxonomy_name'] = organism.get('taxonomy_name')
                # bios['taxon_id'] = int(organism.get('taxonomy_id'))
                # bios['attribute_recs'] = []
                # bios['attributes'] = []
                # for attribute in elem.findall('.//Attribute'):
                #     attrec = attribute.attrib
                #     attrec['value'] = attribute.text
                #     bios['attribute_recs'].append(attrec)
                #     try:
                #         bios['attributes'].append(attrec['harmonized_name'])
                #     except:
                #         bios['attributes'].append(attrec['attribute_name'])

                # bios['model'] = elem.findtext('.//Model')
                # #print(json.dumps(bios))
                # #res = es.index(index="bioes", doc_type='biosample', id=bios['id'], body=bios)
                # elem.clear()
                # return bios



import aiohttp
import asyncio
import click
import ujson as json


class Zooma(object):
    def __init__(self):
        self.url = "https://www.ebi.ac.uk/spot/zooma/v2/api/services/"


    async def annotate(self,
                       property_value: str,
                       property_type: str = None) -> dict:
        params = {"propertyValue": property_value}
   
        if(property_type is not None):
            params["propertyType"] = property_type
        async with aiohttp.ClientSession() as session:
            async with session.get(self.url+"annotate", params=params) as resp:
                return await resp.json()



if __name__ == '__main__':
    z = Zooma()
    loop = asyncio.get_event_loop()
    loop.run_until_complete(z.annotate("mus musculus"))



import omicidx.geo.parser as op
import click
import json


@click.group()
def cli():
    pass


@cli.command()
@click.option('--gse', help='A single GSE accession')
def gse_to_json(gse):
    print(op.gse_to_json(gse))


@cli.command()
@click.option('--term', help='something like "2019/01/28:2019/02/28[PDAT]"')
@click.option("--outfile",
              type=click.File('w'),
              help="output file name, default is stdout")
def bulk_gse_to_json(term, outfile='/dev/stdout'):
    op.bulk_gse_to_json(term, outfile)


@cli.command()
@click.option('--term', help='something like "2019/01/28:2019/02/28[PDAT]"')
@click.option("--outfile", help="output file name, default is stdout")
def gse_accessions(term, outfile='/dev/stdout'):
    op.gse_accessions(term, outfile)


if __name__ == '__main__':
    cli()



from .sra_parsers import *
import logging
import boto3
import decimal
from boto3.dynamodb.conditions import Key

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger('lambda_handlers')
logger.setLevel(logging.INFO)

dynamodb = boto3.resource('dynamodb')
experiment_table = dynamodb.Table("sra_experiment")
run_table = dynamodb.Table("sra_run")
sample_table = dynamodb.Table("sra_sample")
study_table = dynamodb.Table("sra_study")


def replace_floats(obj):
    if isinstance(obj, list):
        for i in range(len(obj)):
            obj[i] = replace_floats(obj[i])
        return obj
    elif isinstance(obj, dict):
        for k, v in obj.items():
            obj[k] = replace_floats(v)
        return obj
    elif isinstance(obj, float):
        return decimal.Decimal(str(obj))
    else:
        return obj


def replace_decimals(obj):
    if isinstance(obj, list):
        for i in range(len(obj)):
            obj[i] = replace_decimals(obj[i])
        return obj
    elif isinstance(obj, dict):
        for k, v in obj.items():
            obj[k] = replace_decimals(v)
        return obj
    elif isinstance(obj, decimal.Decimal):
        if obj % 1 == 0:
            return int(obj)
        else:
            return float(obj)
    else:
        return obj


def full_experiment_package(accession):
    # note that this returns a list right now, as we can use any
    # accession type. In general, use SRXNNNNNN to get a list
    # of length 1.
    v = load_experiment_xml_by_accession(accession)
    s = list([
        SRAExperimentPackage(exptpkg).data
        for exptpkg in v.getroot().findall(".//EXPERIMENT_PACKAGE")
    ])
    ret = []
    for expt in s:
        for run_idx in range(len(expt['runs'])):
            expt['runs'][run_idx] = run_from_runbrowser(
                expt['runs'][run_idx]['accession'])
            expt['run_count'] = len(expt['runs'])
    return s


def create_dynamodb_table(tname):
    table = dynamodb.create_table(
        TableName=tname,
        KeySchema=[{
            'AttributeName': 'accession',
            'KeyType': 'HASH'  #Partition key
        }],
        AttributeDefinitions=[{
            'AttributeName': 'accession',
            'AttributeType': 'S'
        }],
        ProvisionedThroughput={
            'ReadCapacityUnits': 20,
            'WriteCapacityUnits': 20
        })
    return table


def insert_to_dynamo(expt_pkg):
    for expt in expt_pkg:
        logger.info('inserting {} to dynamodb'.format(
            expt['experiment']['accession']))
        with run_table.batch_writer() as batch:
            for run in expt['runs']:
                batch.put_item(Item=replace_floats(run))
                study_table.put_item(Item=replace_floats(expt['study']))
                sample_table.put_item(Item=replace_floats(expt['sample']))
        del (expt['runs'])
        del (expt['sample'])
        del (expt['study'])
        logger.info("inserted runs from {} to dynamodb".format(
            expt['experiment']['accession']))
        experiment_table.put_item(Item=replace_floats(expt['experiment']))
        logger.info("inserted experiment from {} to dynamodb".format(
            expt['experiment']['accession']))


def lambda_return_full_experiment_json(event, context):
    accession = event['accession']
    logger.info('processing accession: {}'.format(accession))
    # munge experiment package
    return full_experiment_package(accession)


def insert_if_empty(accession):
    logger.info('checking for accession {} in dynamodb'.format(accession))
    response = experiment_table.query(
        KeyConditionExpression=Key('accession').eq(accession))
    if (len(response['Items']) == 0):
        expt_pkg = full_experiment_package(accession)
        logger.info('expt_pkg for accession {} obtained'.format(accession))
        insert_to_dynamo(expt_pkg)
        return (True)
    logger.info("skipped: {}".format(accession))


def lambda_insert_to_dynamo(event, context):
    accession = event['accession']
    logger.info('processing accession: {}'.format(accession))
    insert_if_empty(accession)
    return accession


def lambda_insert_from_sqs_to_dynamo(event, context):
    accession = event['Records'][0]['body']
    logger.info('processing accession: {}'.format(accession))
    insert_if_empty(accession)
    return accession


def main():
    import pandas as pd
    y = pd.read_csv('~/top10k_experiments')
    from multiprocessing.dummy import Pool as ThreadPool
    pool = ThreadPool(9)
    pool.map(lambda row: insert_if_empty(row), y.itertuples())


if __name__ == '__main__':
    main()



# SRP006387|*Legume|C0023263|140062|ET|Entry Term Replacement for "Fabaceae";RtM via: Fabaceae;checkForceMH Boosted;Forced Non-Leaf Node Lookup:legumes|TI|MM;RC|181^6^0
# SRP006387|*High-Throughput Nucleotide Sequencing|C2936621|119455|MH|RtM via: High-Throughput Nucleotide Sequencing;RtM via: Massively-Parallel Sequencing;Forced Lookup:next generation sequencing|TI|MM;RC|1475^15^1;1492^10^1;1355^26^0;94^26^0

from pydantic import BaseModel
from typing import List


class MTIResult(BaseModel):
    identifier: str
    name: str
    main: bool = False
    concept: str
    score: int = None
    entry_type: List[str]
    explanation: List[str]
    location: List[str]
    stuff: List[str]
    occurrences: List[str]

    @staticmethod
    def parse_mti_line(line):
        def do_split(s):
            s = s.split(';')
            return s

        line = line.strip()
        line = line.split('|')
        for n in range(0, 9 - len(line)):
            line.append('')

        try:
            mti_result = MTIResult(identifier=line[0],
                                   name=line[1].strip('*'),
                                   main=line[1].startswith('*'),
                                   concept=line[2],
                                   score=int(line[3]),
                                   entry_type=do_split(line[4]),
                                   explanation=do_split(line[5]),
                                   location=do_split(line[6]),
                                   stuff=do_split(line[7]),
                                   occurrences=do_split(line[8]))
            return mti_result
        except:
            return None

    @staticmethod
    def mti_record_iterator(fname):
        with open(fname, 'r') as fhandle:
            for line in fhandle:
                res = MTIResult.parse_mti_line(line)
                if (res is not None):
                    yield res


def main():
    import sys
    fname = sys.argv[1]
    for rec in MTIResult.mti_record_iterator(fname):
        print(rec.json())


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['omicidx',
 'omicidx.bioportal',
 'omicidx.bioportal.db',
 'omicidx.biosample.ebi',
 'omicidx.db',
 'omicidx.geo',
 'omicidx.mti',
 'omicidx.ontologies',
 'omicidx.schema',
 'omicidx.scripts',
 'omicidx.sra']

package_data = \
{'': ['*']}

install_requires = \
['Click',
 'aiohttp>=3.6.2,<4.0.0',
 'asyncpg>=0.20.1,<0.21.0',
 'asyncpgsa>=0.26.3,<0.27.0',
 'biopython==1.75',
 'boto3>=1.9,<2.0',
 'databases>=0.3.2,<0.4.0',
 'gcsfs>=0.8.0,<0.9.0',
 'prefect>=0.13.1,<0.14.0',
 'pronto>=2.0.1,<3.0.0',
 'psycopg2>=2.8.5,<3.0.0',
 'pydantic',
 'requests>=2.22,<3.0',
 'sd_cloud_utils',
 'sphinx_click>=2.3.2,<3.0.0',
 'sqlalchemy>=1.3,<2.0',
 'ujson>=1.35,<2.0']

entry_points = \
{'console_scripts': ['omicidx_tool = omicidx.scripts.cli:cli']}

setup_kwargs = {
    'name': 'omicidx',
    'version': '1.0.2.0',
    'description': 'The OmicIDX project collects, reprocesses, and then republishes metadata from multiple public genomics repositories. Included are the NCBI SRA, Biosample, and GEO databases. Publication is via the cloud data warehouse platform Bigquery, a set of performant search and retrieval APIs, and a set of json-format files for easy incorporation into other projects.',
    'long_description': '',
    'author': 'Sean Davis',
    'author_email': 'seandavi@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/omicidx/omicidx-parsers',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
